from torchhd.datasets.beijing_air_quality import BeijingAirQuality
from torchhd.datasets.isolet import ISOLET
from torchhd.datasets.european_languages import EuropeanLanguages
from torchhd.datasets.ucihar import UCIHAR
from torchhd.datasets.airfoil_self_noise import AirfoilSelfNoise
from torchhd.datasets.emg_hand_gestures import EMGHandGestures
from torchhd.datasets.pamap import PAMAP
from torchhd.datasets.ccpp import CyclePowerPlant
from torchhd.datasets.dataset import CollectionDataset
from torchhd.datasets.dataset import DatasetFourFold
from torchhd.datasets.dataset import DatasetTrainTest
from torchhd.datasets.abalone import Abalone
from torchhd.datasets.adult import Adult


__all__ = [
    "BeijingAirQuality",
    "ISOLET",
    "EuropeanLanguages",
    "UCIHAR",
    "AirfoilSelfNoise",
    "EMGHandGestures",
    "PAMAP",
    "CyclePowerPlant",
    "CollectionDataset",
    "DatasetFourFold",
    "DatasetTrainTest",
    "Abalone",
    "Adult",
]



import pytest
import torch
import string

from torchhd import structures, functional

seed = 2147483644
seed1 = 2147483643
letters = list(string.ascii_lowercase)


class TestFSA:
    def test_creation_dim(self):
        F = structures.FiniteStateAutomata(10000)
        assert torch.equal(F.value, torch.zeros(10000))

    def test_generator(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv1 = functional.random_hv(60, 10000, generator=generator)

        generator = torch.Generator()
        generator.manual_seed(seed)
        hv2 = functional.random_hv(60, 10000, generator=generator)

        assert (hv1 == hv2).min().item()

    def test_add_transition(self):
        generator = torch.Generator()
        generator1 = torch.Generator()
        generator.manual_seed(seed)
        generator1.manual_seed(seed1)
        tokens = functional.random_hv(10, 10, generator=generator)
        actions = functional.random_hv(10, 10, generator=generator1)

        F = structures.FiniteStateAutomata(10)

        F.add_transition(tokens[0], actions[1], actions[2])
        assert torch.equal(
            F.value,
            torch.tensor([1.0, 1.0, -1.0, 1.0, 1.0, 1.0, -1.0, -1.0, -1.0, 1.0]),
        )
        F.add_transition(tokens[1], actions[1], actions[3])
        assert torch.equal(
            F.value, torch.tensor([0.0, 0.0, -2.0, 2.0, 0.0, 2.0, 0.0, -2.0, -2.0, 0.0])
        )
        F.add_transition(tokens[2], actions[1], actions[3])
        assert torch.equal(
            F.value,
            torch.tensor([1.0, 1.0, -3.0, 1.0, 1.0, 3.0, -1.0, -1.0, -1.0, 1.0]),
        )

    def test_transition(self):
        generator = torch.Generator()
        generator1 = torch.Generator()
        generator.manual_seed(seed)
        generator1.manual_seed(seed1)
        tokens = functional.random_hv(10, 10, generator=generator)
        states = functional.random_hv(10, 10, generator=generator1)

        F = structures.FiniteStateAutomata(10)

        F.add_transition(tokens[0], states[1], states[2])
        F.add_transition(tokens[1], states[1], states[3])
        F.add_transition(tokens[2], states[1], states[5])

        assert (
            torch.argmax(
                functional.cos_similarity(F.transition(states[1], tokens[0]), states)
            ).item()
            == 2
        )
        assert (
            torch.argmax(
                functional.cos_similarity(F.transition(states[1], tokens[1]), states)
            ).item()
            == 3
        )
        assert (
            torch.argmax(
                functional.cos_similarity(F.transition(states[1], tokens[2]), states)
            ).item()
            == 5
        )

    def test_clear(self):
        generator = torch.Generator()
        generator1 = torch.Generator()
        generator.manual_seed(seed)
        generator1.manual_seed(seed1)
        tokens = functional.random_hv(10, 10, generator=generator)
        states = functional.random_hv(10, 10, generator=generator1)

        F = structures.FiniteStateAutomata(10)

        F.add_transition(tokens[0], states[1], states[2])
        F.add_transition(tokens[1], states[1], states[3])
        F.add_transition(tokens[2], states[1], states[5])

        F.clear()
        assert torch.equal(
            F.value, torch.tensor([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        )



import math
from typing import Type, Union
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

import torchhd.functional as functional
from torchhd.base import VSA_Model
from torchhd.map import MAP

__all__ = [
    "Identity",
    "Random",
    "Level",
    "Thermometer",
    "Circular",
    "Projection",
    "Sinusoid",
]


class Identity(nn.Embedding):
    """Embedding wrapper around :func:`~torchhd.functional.identity_hv`.

    Class inherits from `Embedding <https://pytorch.org/docs/stable/generated/torch.nn.Embedding.html>`_ and supports the same keyword arguments.

    Args:
        num_embeddings (int): the number of hypervectors to generate.
        embedding_dim (int): the dimensionality of the hypervectors.
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

    Examples::

        >>> emb = embeddings.Identity(5, 3)
        >>> idx = torch.LongTensor([0, 1, 4])
        >>> emb(idx)
        tensor([[1., 1., 1.],
                [1., 1., 1.],
                [1., 1., 1.]])

    """

    def __init__(self, num_embeddings, embedding_dim, requires_grad=False, **kwargs):
        super(Identity, self).__init__(num_embeddings, embedding_dim, **kwargs)
        self.weight.requires_grad = requires_grad

    def reset_parameters(self):
        factory_kwargs = {
            "device": self.weight.data.device,
            "dtype": self.weight.data.dtype,
        }

        self.weight.data.copy_(
            functional.identity_hv(
                self.num_embeddings, self.embedding_dim, **factory_kwargs
            )
        )

        self._fill_padding_idx_with_zero()

    def forward(self, input: Tensor) -> Tensor:
        return super().forward(input).as_subclass(MAP)


class Random(nn.Embedding):
    """Embedding wrapper around :func:`~torchhd.functional.random_hv`.

    Class inherits from `Embedding <https://pytorch.org/docs/stable/generated/torch.nn.Embedding.html>`_ and supports the same keyword arguments.

    Args:
        num_embeddings (int): the number of hypervectors to generate.
        embedding_dim (int): the dimensionality of the hypervectors.
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

    Examples::

        >>> emb = embeddings.Random(5, 3)
        >>> idx = torch.LongTensor([0, 1, 4])
        >>> emb(idx)
        tensor([[ 1., -1.,  1.],
                [ 1., -1.,  1.],
                [ 1.,  1.,  1.]])

    """

    def __init__(self, num_embeddings, embedding_dim, requires_grad=False, **kwargs):
        super(Random, self).__init__(num_embeddings, embedding_dim, **kwargs)
        self.weight.requires_grad = requires_grad

    def reset_parameters(self):
        factory_kwargs = {
            "device": self.weight.data.device,
            "dtype": self.weight.data.dtype,
        }

        self.weight.data.copy_(
            functional.random_hv(
                self.num_embeddings, self.embedding_dim, **factory_kwargs
            )
        )

        self._fill_padding_idx_with_zero()

    def forward(self, input: Tensor) -> Tensor:
        return super().forward(input).as_subclass(MAP)


class Level(nn.Embedding):
    """Embedding wrapper around :func:`~torchhd.functional.level_hv`.

    Class inherits from `Embedding <https://pytorch.org/docs/stable/generated/torch.nn.Embedding.html>`_ and supports the same keyword arguments.

    Args:
        num_embeddings (int): the number of hypervectors to generate.
        embedding_dim (int): the dimensionality of the hypervectors.
        low (float, optional): The lower bound of the real number range that the levels represent. Default: ``0.0``
        high (float, optional): The upper bound of the real number range that the levels represent. Default: ``1.0``
        randomness (float, optional): r-value to interpolate between level at ``0.0`` and random-hypervectors at ``1.0``. Default: ``0.0``.
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

    Examples::

        >>> emb = embeddings.Level(5, 10, low=-1, high=2)
        >>> x = torch.FloatTensor([0.3, 1.9, -0.8])
        >>> emb(x)
        tensor([[-1.,  1.,  1., -1., -1., -1., -1.,  1.,  1.,  1.],
                [-1.,  1., -1.,  1., -1., -1., -1.,  1.,  1., -1.],
                [ 1., -1.,  1., -1., -1., -1.,  1.,  1., -1.,  1.]])

    """

    def __init__(
        self,
        num_embeddings,
        embedding_dim,
        low=0.0,
        high=1.0,
        randomness=0.0,
        requires_grad=False,
        **kwargs
    ):
        self.low_value = low
        self.high_value = high
        self.randomness = randomness

        super(Level, self).__init__(num_embeddings, embedding_dim, **kwargs)
        self.weight.requires_grad = requires_grad

    def reset_parameters(self):
        factory_kwargs = {
            "device": self.weight.data.device,
            "dtype": self.weight.data.dtype,
        }

        self.weight.data.copy_(
            functional.level_hv(
                self.num_embeddings,
                self.embedding_dim,
                randomness=self.randomness,
                **factory_kwargs
            )
        )

        self._fill_padding_idx_with_zero()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        indices = functional.value_to_index(
            input, self.low_value, self.high_value, self.num_embeddings
        ).clamp(0, self.num_embeddings - 1)

        return super(Level, self).forward(indices).as_subclass(MAP)


class Thermometer(nn.Embedding):
    """Embedding wrapper around :func:`~torchhd.functional.thermometer_hv`.

    Class inherits from `Embedding <https://pytorch.org/docs/stable/generated/torch.nn.Embedding.html>`_ and supports the same keyword arguments.

    Args:
        num_embeddings (int): the number of hypervectors to generate.
        embedding_dim (int): the dimensionality of the hypervectors.
        low (float, optional): The lower bound of the real number range that the levels represent. Default: ``0.0``
        high (float, optional): The upper bound of the real number range that the levels represent. Default: ``1.0``
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

    Examples::

        >>> emb = embeddings.Thermometer(11, 10, low=-1, high=2)
        >>> x = torch.FloatTensor([0.3, 1.9, -0.8])
        >>> emb(x)
        tensor([[ 1.,  1.,  1.,  1., -1., -1., -1., -1., -1., -1.],
                [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
                [ 1., -1., -1., -1., -1., -1., -1., -1., -1., -1.]])

    """

    def __init__(
        self,
        num_embeddings,
        embedding_dim,
        low=0.0,
        high=1.0,
        requires_grad=False,
        **kwargs
    ):
        self.low_value = low
        self.high_value = high

        super(Thermometer, self).__init__(num_embeddings, embedding_dim, **kwargs)
        self.weight.requires_grad = requires_grad

    def reset_parameters(self):
        factory_kwargs = {
            "device": self.weight.data.device,
            "dtype": self.weight.data.dtype,
        }

        self.weight.data.copy_(
            functional.thermometer_hv(
                self.num_embeddings, self.embedding_dim, **factory_kwargs
            )
        )

        self._fill_padding_idx_with_zero()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        indices = functional.value_to_index(
            input, self.low_value, self.high_value, self.num_embeddings
        ).clamp(0, self.num_embeddings - 1)

        return super(Thermometer, self).forward(indices).as_subclass(MAP)


class Circular(nn.Embedding):
    """Embedding wrapper around :func:`~torchhd.functional.circular_hv`.

    Class inherits from `Embedding <https://pytorch.org/docs/stable/generated/torch.nn.Embedding.html>`_ and supports the same keyword arguments.

    Args:
        num_embeddings (int): the number of hypervectors to generate.
        embedding_dim (int): the dimensionality of the hypervectors.
        low (float, optional): The lower bound of the real number range that the circular levels represent. Default: ``0.0``
        high (float, optional): The upper bound of the real number range that the circular levels represent. Default: ``2 * pi``
        randomness (float, optional): r-value to interpolate between circular at ``0.0`` and random-hypervectors at ``1.0``. Default: ``0.0``.
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

    Examples::

        >>> emb = embeddings.Circular(5, 10)
        >>> x = torch.FloatTensor([0.0, 3.14, 6.28])
        >>> emb(x)
        tensor([[ 1., -1.,  1., -1.,  1.,  1.,  1.,  1., -1.,  1.],
                [ 1., -1., -1.,  1.,  1.,  1.,  1., -1., -1., -1.],
                [ 1., -1., -1., -1.,  1.,  1.,  1.,  1.,  1., -1.]])

    """

    def __init__(
        self,
        num_embeddings,
        embedding_dim,
        low=0.0,
        high=2 * math.pi,
        randomness=0.0,
        requires_grad=False,
        **kwargs
    ):
        self.low_value = low
        self.high_value = high
        self.randomness = randomness

        super(Circular, self).__init__(num_embeddings, embedding_dim, **kwargs)
        self.weight.requires_grad = requires_grad

    def reset_parameters(self):
        factory_kwargs = {
            "device": self.weight.data.device,
            "dtype": self.weight.data.dtype,
        }

        self.weight.data.copy_(
            functional.circular_hv(
                self.num_embeddings,
                self.embedding_dim,
                randomness=self.randomness,
                **factory_kwargs
            )
        )

        self._fill_padding_idx_with_zero()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        indices = functional.value_to_index(
            input, self.low_value, self.high_value, self.num_embeddings
        ).remainder(self.num_embeddings - 1)

        return super(Circular, self).forward(indices).as_subclass(MAP)


class Projection(nn.Module):
    r"""Embedding using a random projection matrix.

    Implemented based on `A Theoretical Perspective on Hyperdimensional Computing <https://arxiv.org/abs/2010.07426>`_.
    It computes :math:`x \Phi^{\mathsf{T}}` where :math:`\Phi \in \mathbb{R}^{d \times m}` is a matrix whose rows are uniformly sampled at random from the surface of an :math:`d`-dimensional unit sphere.
    This encoding ensures that similarities in the input space are preserved in the hyperspace.

    Args:
        in_features (int): the dimensionality of the input feature vector.
        out_features (int): the dimensionality of the hypervectors.
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.
        dtype (``torch.dtype``, optional): the desired data type of returned tensor. Default: if ``None``, uses a global default (see ``torch.set_default_tensor_type()``).
        device (``torch.device``, optional):  the desired device of returned tensor. Default: if ``None``, uses the current device for the default tensor type (see torch.set_default_tensor_type()). ``device`` will be the CPU for CPU tensor types and the current CUDA device for CUDA tensor types.

    Examples::

        >>> embed = embeddings.Projection(6, 5)
        >>> x = torch.randn(3, 6)
        >>> x
        tensor([[ 0.4119, -0.4284,  1.8022,  0.3715, -1.4563, -0.2842],
                [-0.3772, -1.2664, -1.5173,  1.3317,  0.4707, -1.3362],
                [-1.8142,  0.0274, -1.0989,  0.8193,  0.7619,  0.9181]])
        >>> embed(x).sign()
        tensor([[-1.,  1.,  1.,  1.,  1.],
                [ 1.,  1.,  1.,  1.,  1.],
                [ 1., -1., -1., -1., -1.]])

    """

    __constants__ = ["in_features", "out_features"]
    in_features: int
    out_features: int
    weight: torch.Tensor

    def __init__(
        self, in_features, out_features, requires_grad=False, device=None, dtype=None
    ):
        factory_kwargs = {"device": device, "dtype": dtype}
        super(Projection, self).__init__()
        self.in_features = in_features
        self.out_features = out_features

        self.weight = nn.parameter.Parameter(
            torch.empty((out_features, in_features), **factory_kwargs),
            requires_grad=requires_grad,
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.normal_(self.weight, 0, 1)
        self.weight.data.copy_(F.normalize(self.weight.data))

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return F.linear(input, self.weight).as_subclass(MAP)


class Sinusoid(nn.Module):
    r"""Embedding using a nonlinear random projection

    Implemented based on `Scalable Edge-Based Hyperdimensional Learning System with Brain-Like Neural Adaptation <https://dl.acm.org/doi/abs/10.1145/3458817.3480958>`_.
    It computes :math:`\cos(x \Phi^{\mathsf{T}} + b) \odot \sin(x \Phi^{\mathsf{T}})` where :math:`\Phi \in \mathbb{R}^{d \times m}` is a matrix whose elements are sampled at random from a standard normal distribution and :math:`b \in \mathbb{R}^{d}` is a vectors whose elements are sampled uniformly at random between 0 and :math:`2\pi`.

    Args:
        in_features (int): the dimensionality of the input feature vector.
        out_features (int): the dimensionality of the hypervectors.
        requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.
        dtype (``torch.dtype``, optional): the desired data type of returned tensor. Default: if ``None``, uses a global default (see ``torch.set_default_tensor_type()``).
        device (``torch.device``, optional):  the desired device of returned tensor. Default: if ``None``, uses the current device for the default tensor type (see torch.set_default_tensor_type()). ``device`` will be the CPU for CPU tensor types and the current CUDA device for CUDA tensor types.

    Examples::

        >>> embed = embeddings.Sinusoid(6, 5)
        >>> x = torch.randn(3, 6)
        >>> x
        tensor([[ 0.5043,  0.3161, -0.0938,  0.6134, -0.1280,  0.3647],
                [-0.1907,  1.6468, -0.3242,  0.8614,  0.3332, -0.2055],
                [-0.8662, -1.3861, -0.1577,  0.1321, -0.1157, -2.8928]])
        >>> embed(x)
        tensor([[-0.0555,  0.2292, -0.1833,  0.0301, -0.2416],
                [-0.0725,  0.7042, -0.5644,  0.2235,  0.3603],
                [-0.9021,  0.8899, -0.9802,  0.3565,  0.2367]])

    """

    __constants__ = ["in_features", "out_features"]
    in_features: int
    out_features: int
    weight: torch.Tensor
    bias: torch.Tensor

    def __init__(
        self, in_features, out_features, requires_grad=False, device=None, dtype=None
    ):
        factory_kwargs = {"device": device, "dtype": dtype}
        super(Sinusoid, self).__init__()
        self.in_features = in_features
        self.out_features = out_features

        self.weight = nn.parameter.Parameter(
            torch.empty((out_features, in_features), **factory_kwargs),
            requires_grad=requires_grad,
        )

        self.bias = nn.parameter.Parameter(
            torch.empty((1, out_features), **factory_kwargs),
            requires_grad=requires_grad,
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.normal_(self.weight, 0, 1)
        nn.init.uniform_(self.bias, 0, 2 * math.pi)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        projected = F.linear(input, self.weight)
        output = torch.cos(projected + self.bias) * torch.sin(projected)
        return output.as_subclass(MAP)



import torch
from typing import List, Set
from torch import Tensor


class VSA_Model(Tensor):
    """Base class

    Each model must implement the methods specified on this base class.
    """

    supported_dtypes: Set[torch.dtype]

    @classmethod
    def empty_hv(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        dtype=None,
        device=None,
    ) -> "VSA_Model":
        """Creates hypervectors representing empty sets"""
        raise NotImplementedError

    @classmethod
    def identity_hv(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        dtype=None,
        device=None,
    ) -> "VSA_Model":
        """Creates identity hypervectors for binding"""
        raise NotImplementedError

    @classmethod
    def random_hv(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        dtype=None,
        device=None,
        generator=None,
    ) -> "VSA_Model":
        """Creates random or uncorrelated hypervectors"""
        raise NotImplementedError

    def bundle(self, other: "VSA_Model") -> "VSA_Model":
        """Bundle the hypervector with other"""
        raise NotImplementedError

    def multibundle(self) -> "VSA_Model":
        """Bundle multiple hypervectors"""
        if self.dim() < 2:
            class_name = self.__class__.__name__
            raise RuntimeError(
                f"{class_name} needs to have at least two dimensions for multibundle, got size: {tuple(self.shape)}"
            )

        n = self.size(-2)
        if n == 1:
            return self.unsqueeze(-2)

        tensors: List[VSA_Model] = torch.unbind(self, dim=-2)
        print(type(tensors[0]))

        output = tensors[0].bundle(tensors[1])
        for i in range(2, n):
            output = output.bundle(tensors[i])

        return output

    def bind(self, other: "VSA_Model") -> "VSA_Model":
        """Bind the hypervector with other"""
        raise NotImplementedError

    def multibind(self) -> "VSA_Model":
        """Bind multiple hypervectors"""
        if self.dim() < 2:
            class_name = self.__class__.__name__
            raise RuntimeError(
                f"{class_name} data needs to have at least two dimensions for multibind, got size: {tuple(self.shape)}"
            )

        n = self.size(-2)
        if n == 1:
            return self.unsqueeze(-2)

        tensors: List[VSA_Model] = torch.unbind(self, dim=-2)

        output = tensors[0].bind(tensors[1])
        for i in range(2, n):
            output = output.bind(tensors[i])

        return output

    def inverse(self) -> "VSA_Model":
        """Inverse the hypervector for binding"""
        raise NotImplementedError

    def negative(self) -> "VSA_Model":
        """Negate the hypervector for the bundling inverse"""
        raise NotImplementedError

    def permute(self, n: int = 1) -> "VSA_Model":
        """Permute the hypervector"""
        raise NotImplementedError

    def dot_similarity(self, others: "VSA_Model") -> Tensor:
        """Inner product with other hypervectors"""
        raise NotImplementedError

    def cos_similarity(self, others: "VSA_Model") -> Tensor:
        """Cosine similarity with other hypervectors"""
        raise NotImplementedError



import zipfile
import requests
import re
import tqdm

# Code adapted from:
# https://github.com/wkentaro/gdown/blob/941200a9a1f4fd7ab903fb595baa5cad34a30a45/gdown/download.py
# https://stackoverflow.com/questions/38511444/python-download-files-from-google-drive-using-url


def download_file(url, destination):
    response = requests.get(url, allow_redirects=True, stream=True)
    write_response_to_disk(response, destination)


def download_file_from_google_drive(file_id, destination):
    URL = "https://docs.google.com/uc"
    params = dict(id=file_id, export="download")

    with requests.Session() as session:

        response = session.get(URL, params=params, stream=True)

        # downloads right away
        if "Content-Disposition" in response.headers:
            write_response_to_disk(response, destination)
            return

        # try to find a confirmation token
        token = get_google_drive_confirm_token(response)

        if token:
            params = dict(id=id, confirm=token)
            response = session.get(URL, params=params, stream=True)

        # download if confirmation token worked
        if "Content-Disposition" in response.headers:
            write_response_to_disk(response, destination)
            return

        # extract download url from confirmation page
        url = get_url_from_gdrive_confirmation(response.text)
        response = session.get(url, stream=True)

        write_response_to_disk(response, destination)


def get_google_drive_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith("download_warning"):
            return value

    return None


def get_url_from_gdrive_confirmation(contents):
    url = ""
    for line in contents.splitlines():
        m = re.search(r'href="(\/uc\?export=download[^"]+)', line)
        if m:
            url = "https://docs.google.com" + m.groups()[0]
            url = url.replace("&amp;", "&")
            break
        m = re.search('id="downloadForm" action="(.+?)"', line)
        if m:
            url = m.groups()[0]
            url = url.replace("&amp;", "&")
            break
        m = re.search('id="download-form" action="(.+?)"', line)
        if m:
            url = m.groups()[0]
            url = url.replace("&amp;", "&")
            break
        m = re.search('"downloadUrl":"([^"]+)', line)
        if m:
            url = m.groups()[0]
            url = url.replace("\\u003d", "=")
            url = url.replace("\\u0026", "&")
            break
        m = re.search('<p class="uc-error-subcaption">(.*)</p>', line)
        if m:
            error = m.groups()[0]
            raise RuntimeError(error)
    if not url:
        raise RuntimeError(
            "Cannot retrieve the public link of the file. "
            "You may need to change the permission to "
            "'Anyone with the link', or have had many accesses."
        )
    return url


def get_download_progress_bar(response):
    total = response.headers.get("Content-Length")
    if total is not None:
        total = int(total)

    if total is not None:
        pbar = tqdm.tqdm(total=total, unit="B", unit_scale=True)

    def update(progress):
        if total is not None:
            pbar.update(progress)

    return update


def write_response_to_disk(response, destination):
    CHUNK_SIZE = 32768

    update_progress_bar = get_download_progress_bar(response)

    with open(destination, "wb") as file:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                file.write(chunk)
                update_progress_bar(len(chunk))


def unzip_file(file, destination):
    with zipfile.ZipFile(file, "r") as zip_file:
        zip_file.extractall(destination)



import os
import os.path
from typing import Callable, Optional, Tuple, List
import torch
from torch import Tensor
from torch.utils import data
import pandas as pd
import tarfile
import numpy as np

from .utils import download_file_from_google_drive


class CollectionDataset(data.Dataset):
    """Generic class for loading datasets used in `Do we Need Hundreds of Classifiers to Solve Real World Classification Problems? <https://jmlr.org/papers/v15/delgado14a.html>`_.

    Args:
        root (string): Root directory containing the files of the dataset.
        transform (callable, optional): A function/transform that takes in an torch.FloatTensor
            and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        download (bool, optional): If True, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
    """

    # Name of the dataset. Used when extracting the corresponding files from archive; needs to be specified by subclasses
    name: str
    data: Tensor
    targets: Tensor

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ):
        root = os.path.join(root, self.name)
        root = os.path.expanduser(root)
        self.root = root
        os.makedirs(self.root, exist_ok=True)

        self.transform = transform
        self.target_transform = target_transform

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can try to specify download=True to download it"
            )

        self._load_data()

    def _check_integrity(self) -> bool:
        return NotImplemented

    def _load_data(self):
        return NotImplemented

    def __len__(self) -> int:
        return self.data.size(0)

    def __repr__(self) -> str:
        return f"{self.name}({len(self)})"

    def __getitem__(self, index: int) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        """
        Args:
            index (int): Index

        Returns:
            Tuple[torch.FloatTensor, torch.LongTensor]: (sample, target) where target is the index of the target class
        """
        sample = self.data[index]
        label = self.targets[index]

        if self.transform:
            sample = self.transform(sample)

        if self.target_transform:
            label = self.target_transform(label)

        return sample, label

    def download(self):
        """Download the data if it doesn't exist already."""

        if self._check_integrity():
            print("Files are already downloaded and verified")
            return

        # original data url:
        # http://persoal.citius.usc.es/manuel.fernandez.delgado/papers/jmlr/data.tar.gz

        data_dir = os.path.join(self.root, os.pardir)
        archive_path = os.path.join(data_dir, "data_hundreds_classifiers.tar.gz")

        if os.path.isfile(archive_path):
            print("Archive file is already downloaded")
        else:
            download_file_from_google_drive(
                "1Z3tEzCmR-yTvn1ZlAXaeAuVB5a9oCAkk", archive_path
            )

        # Extract only the requested dataset from the archive
        with tarfile.open(archive_path) as file:
            for member in file.getmembers():
                if member.name.startswith(self.name):
                    file.extract(member, data_dir)


class DatasetFourFold(CollectionDataset):
    """Generic class for loading datasets without separate test data that were used in `Do we Need Hundreds of Classifiers to Solve Real World Classification Problems? <https://jmlr.org/papers/v15/delgado14a.html>`_.

    Args:
        root (string): Root directory containing the files of the dataset.
        train (bool, optional): If True, returns training (sub)set from the file storing training data as further determined by fold and hyper_search variables.
            Otherwise returns a subset of train dataset if hypersearch is performed (``hyper_search = True``) if not (``hyper_search = False``) returns a subset of training dataset
            as specified in ``conxuntos_kfold.dat`` if fold number is correct. Otherwise issues an error.
        fold (int, optional): Specifies which fold number to use. The default value of -1 returns all the training data from the corresponding file.
            Values between 0 and 3 specify, which fold in ``conxuntos_kfold.dat`` to use. Relevant only if hyper_search is set to False and ``0 <= fold <= 3``.
            Indices in even rows (zero indexing) of ``conxuntos_kfold.dat`` correspond to train subsets while indices in odd rows correspond to test subsets.
        hyper_search (bool, optional): If True, creates dataset using indeces in ``conxuntos.dat``. This split is used for hyperparameter search. The first row corresponds to train indices (used if ``train = True``)
            while the second row corresponds to test indices (used if ``train = False``).
        transform (callable, optional): A function/transform that takes in an torch.FloatTensor
            and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        download (bool, optional): If True, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
    """

    # Number of folds for cross validation
    num_folds = 4

    def __init__(
        self,
        root: str,
        train: bool = True,
        fold: int = -1,
        hyper_search: bool = False,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ):
        if (fold < -1) or (fold >= self.num_folds):
            raise ValueError(
                f"Fold number {fold} is not available. Fold should be between -1 and {self.num_folds - 1}"
            )

        if not train and fold == -1 and not hyper_search:
            raise ValueError(
                "This dataset does not have a separate file for test data. Please check that fold is specified correctly."
            )

        self.train = train
        self.fold = fold
        self.hyper_search = hyper_search
        super().__init__(root, transform, target_transform, download)

    def _check_integrity(self) -> bool:
        if not os.path.isdir(self.root):
            return False

        # Check if the root directory contains the required files
        has_train_file = os.path.isfile(os.path.join(self.root, self.name + "_R.dat"))

        has_k_fold_file = os.path.isfile(os.path.join(self.root, "conxuntos_kfold.dat"))
        has_fold_file = os.path.isfile(os.path.join(self.root, "conxuntos.dat"))

        if has_train_file and has_k_fold_file and has_fold_file:
            return True

        # TODO: Add more specific checks like an MD5 checksum

        return False

    def _load_data(self):
        data_path = os.path.join(self.root, self.name + "_R.dat")
        data = np.loadtxt(data_path, skiprows=1, dtype=np.float32)
        # Separate the targets from the data
        targets = torch.from_numpy(data[:, -1].astype(np.int64))
        data = torch.from_numpy(data[:, 1:-1])

        # Load fold used in hyperparameter search if necessary
        if self.hyper_search:
            # Files with pre-generated folds with indices
            hyper_split_path = os.path.join(self.root, "conxuntos.dat")
            line_idx = 0 if self.train else 1

            with open(hyper_split_path, "r") as file:
                lines = file.readlines()
                indices = np.fromstring(lines[line_idx], sep=" ", dtype=np.int64)
                indices = torch.from_numpy(indices)

                data = data[indices]
                targets = targets[indices]

        elif self.fold != -1:
            # Files with pre-generated folds with indices
            k_fold_path = os.path.join(self.root, "conxuntos_kfold.dat")
            fold_idx = self.fold * 2
            if not self.train:
                fold_idx += 1

            with open(k_fold_path, "r") as file:
                lines = file.readlines()
                indices = np.fromstring(lines[fold_idx], sep=" ", dtype=np.int64)
                indices = torch.from_numpy(indices)

                data = data[indices]
                targets = targets[indices]

        self.data = data
        self.targets = targets


class DatasetTrainTest(CollectionDataset):
    """Generic class for loading datasets with separate files for train and test data that were used in `Do we Need Hundreds of Classifiers to Solve Real World Classification Problems? <https://jmlr.org/papers/v15/delgado14a.html>`_.

    Args:
        root (string): Root directory containing the files of the dataset.
        train (bool, optional): If True, returns training (sub)set from the file storing training data as further determined by hyper_search variable.
            Otherwise returns a subset of train dataset if hyperparameter search is performed (``hyper_search = True``) if not (``hyper_search = False``) returns test set.
        hyper_search (bool, optional): If True, creates dataset using indices in ``conxuntos.dat``. This split is used for hyperparameter search. The first row corresponds to train indices (used if ``train = True``)
            while the second row corresponds to test indices (used if ``train = False``).
        transform (callable, optional): A function/transform that takes in an torch.FloatTensor
            and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        download (bool, optional): If True, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
    """

    def __init__(
        self,
        root: str,
        train: bool = True,
        hyper_search: bool = False,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ):
        self.train = train
        self.hyper_search = hyper_search
        super().__init__(root, transform, target_transform, download)

    def _check_integrity(self) -> bool:
        if not os.path.isdir(self.root):
            return False

        # Check if the root directory contains the required files
        has_train_file = os.path.isfile(
            os.path.join(self.root, self.name + "_train_R.dat")
        )

        has_k_fold_file = os.path.isfile(os.path.join(self.root, "conxuntos_kfold.dat"))
        has_fold_file = os.path.isfile(os.path.join(self.root, "conxuntos.dat"))

        if has_train_file and has_k_fold_file and has_fold_file:
            return True

        # TODO: Add more specific checks like an MD5 checksum

        return False

    def _load_data(self):
        if self.train or self.hyper_search:
            data_name = self.name + "_train_R.dat"
        else:
            data_name = self.name + "_test_R.dat"

        data_path = os.path.join(self.root, data_name)

        data = np.loadtxt(data_path, skiprows=1, dtype=np.float32)
        # Separate the targets from the data
        targets = torch.from_numpy(data[:, -1].astype(np.int64))
        data = torch.from_numpy(data[:, 1:-1])

        # Load fold used in hyperparameter search if necessary
        if self.hyper_search:

            # Files with pre-generated folds with indices
            hyper_split_path = os.path.join(self.root, "conxuntos.dat")
            line_idx = 0 if self.train else 1

            with open(hyper_split_path, "r") as file:
                lines = file.readlines()
                indices = np.fromstring(lines[line_idx], sep=" ", dtype=np.int64)
                indices = torch.from_numpy(indices)

                data = data[indices]
                targets = targets[indices]

        self.data = data
        self.targets = targets



import pytest
import torch
import string

from torchhd import structures, functional

seed = 2147483644
letters = list(string.ascii_lowercase)


class TestMemory:
    def test_creation(self):
        M = structures.Memory()

        assert M.keys == []
        assert M.values == []

    def test_generator(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv1 = functional.random_hv(60, 10000, generator=generator)

        generator = torch.Generator()
        generator.manual_seed(seed)
        hv2 = functional.random_hv(60, 10000, generator=generator)

        assert (hv1 == hv2).min().item()

    def test_add(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        keys_hv = functional.random_hv(len(letters), 10000, generator=generator)

        M = structures.Memory()
        M.add(keys_hv[0], letters[0])
        M.add(keys_hv[1], letters[1])
        M.add(keys_hv[2], letters[2])

        assert torch.equal(M.keys[0], keys_hv[0])
        assert torch.equal(M.keys[1], keys_hv[1])
        assert torch.equal(M.keys[2], keys_hv[2])
        assert M.values[0] == letters[0]
        assert M.values[1] == letters[1]
        assert M.values[2] == letters[2]

    def test_index(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        keys_hv = functional.random_hv(len(letters), 10000, generator=generator)

        M = structures.Memory()
        M.add(keys_hv[0], letters[0])
        M.add(keys_hv[1], letters[1])
        M.add(keys_hv[2], letters[2])

        assert M.index(keys_hv[0]) == 0
        assert M.index(keys_hv[1]) == 1
        assert M.index(keys_hv[2]) == 2

    def test_length(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        keys_hv = functional.random_hv(len(letters), 10000, generator=generator)

        M = structures.Memory()
        M.add(keys_hv[0], letters[0])
        M.add(keys_hv[1], letters[1])
        M.add(keys_hv[2], letters[2])

        assert len(M) == 3
        del M[keys_hv[0]]

        assert len(M) == 2

        M.add(keys_hv[0], letters[0])
        assert len(M) == 3

    def test_getitem(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        keys_hv = functional.random_hv(len(letters), 10000, generator=generator)

        M = structures.Memory()
        M.add(keys_hv[0], letters[0])
        M.add(keys_hv[1], letters[1])
        M.add(keys_hv[2], letters[2])

        assert M[keys_hv[0]][1] == letters[0]
        assert M[keys_hv[1]][1] == letters[1]
        assert M[keys_hv[2]][1] == letters[2]

    def test_setitem(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        keys_hv = functional.random_hv(len(letters), 10000, generator=generator)

        M = structures.Memory()
        M.add(keys_hv[0], letters[0])
        M.add(keys_hv[1], letters[1])
        M.add(keys_hv[2], letters[2])

        assert len(M) == 3
        assert M[keys_hv[0]][1] == letters[0]
        assert M[keys_hv[1]][1] == letters[1]
        assert M[keys_hv[2]][1] == letters[2]

        M[keys_hv[0]] = letters[3]
        assert len(M) == 3
        assert M[keys_hv[0]][1] == letters[3]

    def test_delitem(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        keys_hv = functional.random_hv(len(letters), 10000, generator=generator)

        M = structures.Memory()
        M.add(keys_hv[0], letters[0])
        M.add(keys_hv[1], letters[1])
        M.add(keys_hv[2], letters[2])

        assert len(M) == 3
        assert M[keys_hv[0]][1] == letters[0]
        assert M[keys_hv[1]][1] == letters[1]
        assert M[keys_hv[2]][1] == letters[2]

        del M[keys_hv[0]]
        try:
            M[keys_hv[0]]
        except IndexError:
            assert True

        assert M[keys_hv[1]][1] == letters[1]
        assert M[keys_hv[2]][1] == letters[2]
        assert len(M) == 2



import pytest
import torch

import torchhd
from torchhd import functional

from ..utils import (
    torch_dtypes,
    supported_dtype,
    vsa_models,
)

seed = 2147483644


class TestEmpty_hv:
    @pytest.mark.parametrize("n", [1, 3, 55])
    @pytest.mark.parametrize("d", [84, 10])
    @pytest.mark.parametrize("model", vsa_models)
    def test_shape(self, n, d, model):
        hv = functional.empty_hv(n, d, model)

        assert hv.dim() == 2
        assert hv.size(0) == n
        assert hv.size(1) == d

    @pytest.mark.parametrize("dtype", torch_dtypes)
    @pytest.mark.parametrize("model", vsa_models)
    def test_value(self, dtype, model):

        if not supported_dtype(dtype, model):
            with pytest.raises(ValueError):
                functional.empty_hv(3, 26, model, dtype=dtype)

            return

        hv = functional.empty_hv(8, 26, model, dtype=dtype)
        assert hv.requires_grad == False
        assert hv.dim() == 2
        assert hv.size(0) == 8
        assert hv.size(1) == 26

        if model == torchhd.BSC:
            assert torch.all((hv == False) | (hv == True)).item()

        else:
            hv = functional.empty_hv(8, 26, model, dtype=dtype)
            assert torch.all(hv == 0.0).item()

    @pytest.mark.parametrize("dtype", torch_dtypes)
    @pytest.mark.parametrize("model", vsa_models)
    def test_device(self, dtype, model):
        if not supported_dtype(dtype, model):
            return

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        hv = functional.empty_hv(3, 52, model, device=device, dtype=dtype)
        assert hv.device == device

    def test_uses_default_dtype(self):
        hv = functional.empty_hv(3, 52, torchhd.BSC)
        assert hv.dtype == torch.bool

        torch.set_default_dtype(torch.float32)
        hv = functional.empty_hv(3, 52, torchhd.MAP)
        assert hv.dtype == torch.float32
        hv = functional.empty_hv(3, 52, torchhd.HRR)
        assert hv.dtype == torch.float32

        torch.set_default_dtype(torch.float64)
        hv = functional.empty_hv(3, 52, torchhd.MAP)
        assert hv.dtype == torch.float64
        hv = functional.empty_hv(3, 52, torchhd.HRR)
        assert hv.dtype == torch.float64

        hv = functional.empty_hv(3, 52, torchhd.FHRR)
        assert hv.dtype == torch.complex64

    def test_requires_grad(self):
        hv = functional.empty_hv(3, 52, torchhd.MAP, requires_grad=True)
        assert hv.requires_grad == True

        hv = functional.empty_hv(3, 52, torchhd.HRR, requires_grad=True)
        assert hv.requires_grad == True

        hv = functional.empty_hv(3, 52, torchhd.FHRR, requires_grad=True)
        assert hv.requires_grad == True



from typing import List
from torchhd.datasets import DatasetTrainTest


class Adult(DatasetTrainTest):
    """`Adult <https://archive.ics.uci.edu/ml/datasets/adult>`_ dataset.

    Args:
        root (string): Root directory containing the files of the dataset.
        train (bool, optional): If True, returns training (sub)set from the file storing training data as further determined by hyper_search variable.
            Otherwise returns a subset of train dataset if hyperparameter search is performed (``hyper_search = True``) if not (``hyper_search = False``) returns test set.
        hyper_search (bool, optional): If True, creates dataset using indices in ``conxuntos.dat``. This split is used for hyperparameter search. The first row corresponds to train indices (used if ``train = True``)
            while the second row corresponds to test indices (used if ``train = False``).
        transform (callable, optional): A function/transform that takes in an torch.FloatTensor
            and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        download (bool, optional): If True, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
    """

    name = "adult"
    classes: List[str] = [
        ">50K",
        "<=50K",
    ]



import os
import os.path
import copy
from typing import Callable, Optional, Tuple, List
import torch
from torch.utils import data
import pandas as pd

from .utils import download_file, unzip_file


class PAMAP(data.Dataset):
    """`PAMAP <https://archive.ics.uci.edu/ml/datasets/pamap2+physical+activity+monitoring>`_ dataset.

    Args:
        root (string): Root directory of dataset.
        subjects (list): List of subjects to be loaded in dataset
        optional (bool): If true optional data of some subjectes will be loaded.
        download (bool, optional): If True, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        transform (callable, optional): A function/transform that takes in an torch.FloatTensor
            and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
    """

    classes: List[str] = [
        "lying",
        "sitting",
        "standing",
        "walking",
        "running",
        "cycling",
        "nordic walking",
        "watching TV",
        "computer work",
        "car driving",
        "ascending stairs",
        "descending stairs",
        "vacuum cleaning",
        "ironing",
        "folding laundry",
        "house cleaning",
        "playing soccer",
        "rope jumping",
    ]

    columns: List[str] = [
        "timestamp",
        "activity",
        "heartRate",
        "handTemp",
        "handAcc11",
        "handAcc12",
        "handAcc13",
        "handAcc21",
        "handAcc22",
        "handAcc23",
        "handGyro1",
        "handGyro2",
        "handGyro3",
        "handMagnetometer1",
        "handMagnetometer2",
        "handMagnetometer3",
        "handOrientation1",
        "handOrientation2",
        "handOrientation3",
        "handOrientation4",
        "chestTemp",
        "chestAcc11",
        "chestAcc12",
        "chestAcc13",
        "chestAcc21",
        "chestAcc22",
        "chestAcc23",
        "chestGyro1",
        "chestGyro2",
        "chestGyro3",
        "chestMagnetometer1",
        "chestMagnetometer2",
        "chestMagnetometer3",
        "chestOrientation1",
        "chestOrientation2",
        "chestOrientation3",
        "chestOrientation4",
        "ankleTemp",
        "ankleAcc11",
        "ankleAcc12",
        "ankleAcc13",
        "ankleAcc21",
        "ankleAcc22",
        "ankleAcc23",
        "ankleGyro1",
        "ankleGyro2",
        "ankleGyro3",
        "ankleMagnetometer1",
        "ankleMagnetometer2",
        "ankleMagnetometer3",
        "ankleOrientation1",
        "ankleOrientation2",
        "ankleOrientation3",
        "ankleOrientation4",
    ]

    subjects_with_optional_data: List[int] = [0, 4, 5, 7, 8]

    def __init__(
        self,
        root: str,
        subjects: list = [0, 1, 2, 3, 4, 5, 6, 7, 8],
        optional: bool = False,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ):
        root = os.path.join(root, "pamap")
        root = os.path.expanduser(root)
        self.root = root
        self.subjects = subjects
        self.optional = optional
        os.makedirs(self.root, exist_ok=True)

        self.transform = transform
        self.target_transform = target_transform

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it"
            )

        self._load_data()

    def __len__(self) -> int:
        return self.data.size(0)

    def __getitem__(self, index: int) -> Tuple[torch.FloatTensor, torch.LongTensor]:
        """
        Args:
            index (int): Index

        Returns:
            Tuple[torch.FloatTensor, torch.LongTensor]: (sample, target) where target is the index of the target class
        """
        sample = self.data[index]
        label = self.targets[index]

        if self.transform:
            sample = self.transform(sample)

        if self.target_transform:
            label = self.target_transform(label)

        return sample, label

    def _check_integrity(self) -> bool:
        if not os.path.isdir(self.root):
            return False
        # Check if the root directory contains the required files
        has_all_files = []
        for i in [1, 5, 6, 8, 9]:
            has_all_files.append(
                os.path.isfile(
                    os.path.join(
                        self.root, "PAMAP2_Dataset/Optional/subject10" + str(i) + ".dat"
                    )
                )
            )
        for i in range(1, 10):
            has_all_files.append(
                os.path.isfile(
                    os.path.join(
                        self.root, "PAMAP2_Dataset/Protocol/subject10" + str(i) + ".dat"
                    )
                )
            )

        if all(has_all_files):
            return True

        # TODO: Add more specific checks like an MD5 checksum

        return False

    def _load_data(self):
        clean_labels = torch.empty(0, dtype=torch.long)
        clean_features = torch.empty(0, dtype=torch.long)
        for i in self.subjects:
            data = pd.read_csv(
                os.path.join(
                    self.root, "PAMAP2_Dataset/Protocol/subject10" + str(i + 1) + ".dat"
                ),
                delimiter=" ",
                header=None,
            )
            # Adding optional data if requested and exists
            if self.optional and i in self.subjects_with_optional_data:
                optional_data = pd.read_csv(
                    os.path.join(
                        self.root,
                        "PAMAP2_Dataset/Optional/subject10" + str(i + 1) + ".dat",
                    ),
                    delimiter=" ",
                    header=None,
                )
                data = pd.concat([data, optional_data])
            # Activity with value 0 should be discarded in any kind of analysis
            data = data[data[1] != 0]
            cols = copy.copy(self.columns)
            data.columns = cols
            cols.remove("heartRate")
            # Drop Nan values that are not heartRate
            data = data.dropna(subset=cols)
            # Replace Nan values of heart rate for value before
            data.ffill(inplace=True)
            data = data.dropna()
            data = data.reset_index(drop=True)
            activities = data["activity"]
            # Replace data activity lables
            activities.replace(
                [9, 10, 11, 12, 13, 16, 17, 18, 19, 20, 24],
                [8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18],
                inplace=True,
            )
            labels = torch.tensor(activities.values, dtype=torch.long)
            del data["activity"]
            features = torch.tensor(data.values, dtype=torch.long)
            clean_labels = torch.cat((clean_labels, labels))
            clean_features = torch.cat((clean_features, features))

        self.data = clean_features
        self.targets = clean_labels

    def download(self):
        """Download the data if it doesn't exist already."""

        if self._check_integrity():
            print("Files already downloaded and verified")
            return

        zip_file_path = os.path.join(self.root, "data.zip")
        download_file(
            "https://archive.ics.uci.edu/ml/machine-learning-databases/00231/PAMAP2_Dataset.zip",
            zip_file_path,
        )

        unzip_file(zip_file_path, self.root)
        os.remove(zip_file_path)



import pytest
import torch

import torchhd
from torchhd import functional

from ..utils import torch_dtypes, supported_dtype, vsa_models

seed = 2147483644


class TestCircular_hv:
    @pytest.mark.parametrize("n", [1, 3, 55])
    @pytest.mark.parametrize("d", [84, 10])
    @pytest.mark.parametrize("model", vsa_models)
    def test_shape(self, n, d, model):
        hv = functional.circular_hv(n, d, model)

        assert hv.dim() == 2
        assert hv.size(0) == n
        assert hv.size(1) == d

    @pytest.mark.parametrize("model", vsa_models)
    def test_generator(self, model):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv1 = functional.circular_hv(20, 10000, model, generator=generator)

        generator = torch.Generator()
        generator.manual_seed(seed)

        hv2 = functional.circular_hv(20, 10000, model, generator=generator)
        assert torch.all(hv1 == hv2).item()

    @pytest.mark.parametrize("dtype", torch_dtypes)
    @pytest.mark.parametrize("model", vsa_models)
    def test_value(self, dtype, model):

        if not supported_dtype(dtype, model):
            with pytest.raises(ValueError):
                functional.circular_hv(3, 26, model, dtype=dtype)

            return

        generator = torch.Generator()
        generator.manual_seed(seed)

        hv = functional.circular_hv(50, 26000, model, dtype=dtype, generator=generator)
        assert hv.requires_grad == False
        assert hv.dim() == 2
        assert hv.size(0) == 50
        assert hv.size(1) == 26000

        if model == torchhd.BSC:
            assert torch.all((hv == False) | (hv == True)).item()

        elif model == torchhd.MAP:
            assert torch.all((hv == -1) | (hv == 1)).item()

        elif model == torchhd.HRR:
            std, mean = torch.std_mean(hv)
            assert torch.allclose(
                mean, torch.tensor(0.0, dtype=mean.dtype), atol=0.0001
            )

        elif model == torchhd.FHRR:
            mag = hv.abs()
            assert torch.allclose(mag, torch.tensor(1.0, dtype=mag.dtype))

        hv = functional.circular_hv(8, 1000000, model, generator=generator, dtype=dtype)
        sims = functional.cos_similarity(hv[0], hv)
        sims_diff = sims[:-1] - sims[1:]
        assert torch.all(
            sims_diff.sign() == torch.tensor([1, 1, 1, 1, -1, -1, -1])
        ), "second half must get more similar"

        assert torch.allclose(
            sims_diff.abs(), torch.tensor(0.25, dtype=sims_diff.dtype), atol=0.005
        ), "similarity decreases linearly"

    @pytest.mark.parametrize("sparsity", [0.0, 0.1, 0.756, 1.0])
    @pytest.mark.parametrize("dtype", torch_dtypes)
    def test_sparsity(self, sparsity, dtype):
        if not supported_dtype(dtype, torchhd.BSC):
            return

        generator = torch.Generator()
        generator.manual_seed(seed)

        hv = functional.circular_hv(
            1000,
            10000,
            torchhd.BSC,
            generator=generator,
            dtype=dtype,
            sparsity=sparsity,
        )

        calc_sparsity = torch.sum(hv == False).div(10000 * 1000)
        assert torch.allclose(calc_sparsity, torch.tensor(sparsity), atol=0.005)

    @pytest.mark.parametrize("dtype", torch_dtypes)
    def test_device(self, dtype):
        if not supported_dtype(dtype):
            return

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        hv = functional.circular_hv(3, 52, device=device, dtype=dtype)
        assert hv.device == device

    @pytest.mark.parametrize("dtype", torch_dtypes)
    @pytest.mark.parametrize("model", vsa_models)
    def test_device(self, dtype, model):
        if not supported_dtype(dtype, model):
            return

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        hv = functional.circular_hv(3, 52, model, device=device, dtype=dtype)
        assert hv.device == device

    def test_uses_default_dtype(self):
        hv = functional.circular_hv(3, 52, torchhd.BSC)
        assert hv.dtype == torch.bool

        torch.set_default_dtype(torch.float32)
        hv = functional.circular_hv(3, 52, torchhd.MAP)
        assert hv.dtype == torch.float32
        hv = functional.circular_hv(3, 52, torchhd.HRR)
        assert hv.dtype == torch.float32

        torch.set_default_dtype(torch.float64)
        hv = functional.circular_hv(3, 52, torchhd.MAP)
        assert hv.dtype == torch.float64
        hv = functional.circular_hv(3, 52, torchhd.HRR)
        assert hv.dtype == torch.float64

        hv = functional.circular_hv(3, 52, torchhd.FHRR)
        assert hv.dtype == torch.complex64

    def test_requires_grad(self):
        hv = functional.circular_hv(3, 52, torchhd.MAP, requires_grad=True)
        assert hv.requires_grad == True

        hv = functional.circular_hv(3, 52, torchhd.HRR, requires_grad=True)
        assert hv.requires_grad == True

        hv = functional.circular_hv(3, 52, torchhd.FHRR, requires_grad=True)
        assert hv.requires_grad == True



import pytest
import torch
import string

from torchhd import structures, functional

seed = 2147483644
letters = list(string.ascii_lowercase)


class TestTree:
    def test_creation_dim(self):
        T = structures.Tree(10000)
        assert torch.equal(T.value, torch.zeros(10000))

    def test_generator(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv1 = functional.random_hv(60, 10000, generator=generator)

        generator = torch.Generator()
        generator.manual_seed(seed)
        hv2 = functional.random_hv(60, 10000, generator=generator)

        assert (hv1 == hv2).min().item()

    def test_add_leaf(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv = functional.random_hv(len(letters), 10000, generator=generator)
        T = structures.Tree(10000)
        T.add_leaf(hv[0], ["l", "l"])
        assert (
            torch.argmax(functional.cos_similarity(T.get_leaf(["l", "l"]), hv)).item()
            == 0
        )
        T.add_leaf(hv[1], ["l", "r"])
        assert (
            torch.argmax(functional.cos_similarity(T.get_leaf(["l", "r"]), hv)).item()
            == 1
        )

    def test_get_leaf(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv = functional.random_hv(len(letters), 10000, generator=generator)
        T = structures.Tree(10000)
        T.add_leaf(hv[0], ["l", "l"])
        assert (
            torch.argmax(functional.cos_similarity(T.get_leaf(["l", "l"]), hv)).item()
            == 0
        )
        T.add_leaf(hv[1], ["l", "r"])
        assert (
            torch.argmax(functional.cos_similarity(T.get_leaf(["l", "r"]), hv)).item()
            == 1
        )

    def test_clear(self):
        generator = torch.Generator()
        generator.manual_seed(seed)
        hv = functional.random_hv(8, 10, generator=generator)
        T = structures.Tree(10)

        T.add_leaf(hv[0], ["l", "l"])
        T.add_leaf(hv[1], ["l", "r"])

        T.clear()
        assert torch.equal(
            T.value, torch.tensor([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        )



import os
import os.path as path
from typing import Callable, Optional, List
import torch
from torch.utils import data
import numpy as np
import pandas as pd

from .utils import download_file, unzip_file


class UCIHAR(data.Dataset):
    """`UCI Human Activity Recognition <https://archive.ics.uci.edu/ml/datasets/Human+Activity+Recognition+Using+Smartphones>`_ dataset.
    As found in the paper `"Human Activity Recognition Using Smartphones" <https://ieeexplore.ieee.org/document/8567275>`_.

    Args:
        root (string): Root directory of dataset where the training and testing samples are located.
        train (bool, optional): If True, creates dataset from UCIHAR-training data,
            otherwise from UCIHAR-testing data
        download (bool, optional): If True, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        transform (callable, optional): A function/transform that takes in an torch.LongTensor
            and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
    """

    classes: List[str] = [
        "WALKING",
        "WALKING_UPSTAIRS",
        "WALKING_DOWNSTAIRS",
        "SITTING",
        "STANDING",
        "LAYING",
    ]

    def __init__(
        self,
        root: str,
        train: bool = True,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ):
        root = path.join(root, "ucihar")
        root = os.path.expanduser(root)
        self.root = root
        os.makedirs(self.root, exist_ok=True)

        self.train = train
        self.transform = transform
        self.target_transform = target_transform

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it"
            )

        self._load_data()

    def __len__(self) -> int:
        return self.targets.size(0)

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            Tuple[torch.FloatTensor, torch.LongTensor]: (sample, target) where target is the index of the target class
        """
        sample = self.data[index]
        target = self.targets[index]

        if self.transform:
            sample = self.transform(sample)

        if self.target_transform:
            target = self.target_transform(target)

        return sample, target

    def _check_integrity(self) -> bool:
        if not os.path.isdir(self.root):
            return False

        train_dir = os.path.join(self.root, "train")
        has_train_dir = os.path.isdir(train_dir)
        test_dir = os.path.join(self.root, "test")
        has_test_dir = os.path.isdir(test_dir)

        if not has_train_dir and not has_test_dir:
            return False

        has_train_x = os.path.isfile(os.path.join(train_dir, "X_train.txt"))
        has_train_y = os.path.isfile(os.path.join(train_dir, "y_train.txt"))

        if not has_train_x and not has_train_y:
            return False

        has_test_x = os.path.isfile(os.path.join(test_dir, "X_test.txt"))
        has_test_y = os.path.isfile(os.path.join(train_dir, "y_test.txt"))

        if not has_test_x or not has_test_y:
            return False

        return True

    def _load_data(self):
        data_dir = os.path.join(self.root, "train" if self.train else "test")
        data_file = "X_train.txt" if self.train else "X_test.txt"
        target_file = "y_train.txt" if self.train else "y_test.txt"

        data = pd.read_csv(
            os.path.join(data_dir, data_file), delim_whitespace=True, header=None
        )
        targets = np.loadtxt(
            path.join(data_dir, target_file), delimiter="\n", dtype="int64"
        ).tolist()

        self.data = torch.tensor(data.values, dtype=torch.float)
        self.targets = torch.tensor(targets, dtype=torch.long) - 1

    def download(self):
        """Downloads the dataset if it doesn't exist already"""

        if self._check_integrity():
            print("Files already downloaded and verified")
            return

        zip_file_path = os.path.join(self.root, "data.zip")
        download_file(
            "https://archive.ics.uci.edu/ml/machine-learning-databases/00240/UCI%20HAR%20Dataset.zip",
            zip_file_path,
        )

        unzip_file(zip_file_path, self.root)
        os.remove(zip_file_path)

        source_dir = os.path.join(self.root, "UCI HAR Dataset")
        data_files = os.listdir(source_dir)
        for filename in data_files:
            os.rename(
                os.path.join(source_dir, filename), os.path.join(self.root, filename)
            )

        os.rmdir(source_dir)



import torch
from typing import Set
from torch import Tensor
import torch.nn.functional as F

from torchhd.base import VSA_Model


def biggest_power_two(n):
    """Returns the biggest power of two <= n"""
    # if n is a power of two simply return it
    if not (n & (n - 1)):
        return n

    # else set only the most significant bit
    return int("1" + (len(bin(n)) - 3) * "0", 2)


class BSC(VSA_Model):
    """Binary Spatter Codes

    Proposed in `Binary spatter-coding of ordered K-tuples <https://link.springer.com/chapter/10.1007/3-540-61510-5_146>`_, this model works with binary valued hypervectors.
    """

    supported_dtypes: Set[torch.dtype] = {
        torch.float32,
        torch.float64,
        torch.uint8,
        torch.int8,
        torch.int16,
        torch.int32,
        torch.int64,
        torch.bool,
    }

    @classmethod
    def empty_hv(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        generator=None,
        dtype=torch.bool,
        device=None,
        requires_grad=False,
    ) -> "BSC":
        """Creates a set of hypervectors representing empty sets.

        When bundled with a random-hypervector :math:`x`, the result is :math:`\sim x`.
        Because of the low precession of the BSC model an empty set cannot be explicitly represented, therefore the returned hypervectors are identical to random-hypervectors.

        Args:
            num_vectors (int): the number of hypervectors to generate.
            dimensions (int): the dimensionality of the hypervectors.
            generator (``torch.Generator``, optional): a pseudorandom number generator for sampling.
            dtype (``torch.dtype``, optional): the desired data type of returned tensor. Default: if ``None`` depends on VSA_Model.
            device (``torch.device``, optional):  the desired device of returned tensor. Default: if ``None``, uses the current device for the default tensor type (see torch.set_default_tensor_type()). ``device`` will be the CPU for CPU tensor types and the current CUDA device for CUDA tensor types.
            requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

        Examples::

            >>> torchhd.BSC.empty_hv(3, 6)
            tensor([[False, False, False, False,  True,  True],
                    [False,  True, False, False,  True,  True],
                    [ True, False,  True,  True, False, False]])

            >>> torchhd.BSC.empty_hv(3, 6, dtype=torch.long)
            tensor([[0, 1, 0, 1, 0, 1],
                    [0, 0, 1, 1, 0, 1],
                    [0, 1, 1, 0, 1, 1]])

        """

        if dtype not in cls.supported_dtypes:
            name = cls.__name__
            options = ", ".join([str(x) for x in cls.supported_dtypes])
            raise ValueError(f"{name} vectors must be one of dtype {options}.")

        size = (num_vectors, dimensions)
        result = torch.empty(size, dtype=dtype, device=device)
        result.bernoulli_(0.5, generator=generator)
        result.requires_grad = requires_grad
        return result.as_subclass(cls)

    @classmethod
    def identity_hv(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        dtype=torch.bool,
        device=None,
        requires_grad=False,
    ) -> "BSC":
        """Creates a set of identity hypervectors.

        When bound with a random-hypervector :math:`x`, the result is :math:`x`.

        Args:
            num_vectors (int): the number of hypervectors to generate.
            dimensions (int): the dimensionality of the hypervectors.
            dtype (``torch.dtype``, optional): the desired data type of returned tensor. Default: if ``None`` depends on VSA_Model.
            device (``torch.device``, optional):  the desired device of returned tensor. Default: if ``None``, uses the current device for the default tensor type (see torch.set_default_tensor_type()). ``device`` will be the CPU for CPU tensor types and the current CUDA device for CUDA tensor types.
            requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

        Examples::

            >>> torchhd.BSC.identity_hv(3, 6)
            tensor([[False, False, False, False, False, False],
                    [False, False, False, False, False, False],
                    [False, False, False, False, False, False]])

            >>> torchhd.BSC.identity_hv(3, 6, dtype=torch.long)
            tensor([[0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0]])

        """

        if dtype not in cls.supported_dtypes:
            name = cls.__name__
            options = ", ".join([str(x) for x in cls.supported_dtypes])
            raise ValueError(f"{name} vectors must be one of dtype {options}.")

        result = torch.zeros(
            num_vectors,
            dimensions,
            dtype=dtype,
            device=device,
            requires_grad=requires_grad,
        )
        return result.as_subclass(cls)

    @classmethod
    def random_hv(
        cls,
        num_vectors: int,
        dimensions: int,
        *,
        sparsity=0.5,
        generator=None,
        dtype=torch.bool,
        device=None,
        requires_grad=False,
    ) -> "BSC":
        """Creates a set of random independent hypervectors.

        The resulting hypervectors are sampled uniformly at random from the ``dimensions``-dimensional hyperspace.

        Args:
            num_vectors (int): the number of hypervectors to generate.
            dimensions (int): the dimensionality of the hypervectors.
            sparsity (float, optional): the expected fraction of elements to be in-active. Has no effect on complex hypervectors. Default: ``0.5``.
            generator (``torch.Generator``, optional): a pseudorandom number generator for sampling.
            dtype (``torch.dtype``, optional): the desired data type of returned tensor. Default: if ``None`` depends on VSA_Model.
            device (``torch.device``, optional):  the desired device of returned tensor. Default: if ``None``, uses the current device for the default tensor type (see torch.set_default_tensor_type()). ``device`` will be the CPU for CPU tensor types and the current CUDA device for CUDA tensor types.
            requires_grad (bool, optional): If autograd should record operations on the returned tensor. Default: ``False``.

        Examples::

            >>> torchhd.BSC.random_hv(3, 6)
            tensor([[ True, False, False, False,  True,  True],
                    [False, False,  True, False, False, False],
                    [False, False,  True,  True, False, False]])

            >>> torchhd.BSC.random_hv(3, 6, sparsity=0.1)
            tensor([[ True,  True,  True,  True,  True,  True],
                    [False,  True,  True,  True,  True,  True],
                    [ True,  True,  True,  True,  True,  True]])

            >>> torchhd.BSC.random_hv(3, 6, dtype=torch.long)
            tensor([[1, 1, 0, 0, 0, 1],
                    [0, 1, 0, 0, 1, 1],
                    [0, 1, 1, 0, 0, 0]])

        """
        if dtype not in cls.supported_dtypes:
            name = cls.__name__
            options = ", ".join([str(x) for x in cls.supported_dtypes])
            raise ValueError(f"{name} vectors must be one of dtype {options}.")

        size = (num_vectors, dimensions)
        result = torch.empty(size, dtype=dtype, device=device)
        result.bernoulli_(1.0 - sparsity, generator=generator)
        result.requires_grad = requires_grad
        return result.as_subclass(cls)

    def bundle(self, other: "BSC", *, generator: torch.Generator = None) -> "BSC":
        r"""Bundle the hypervector with other using majority voting.

        This produces a hypervector maximally similar to both.

        The bundling operation is used to aggregate information into a single hypervector.

        Ties in the majority vote are broken at random. For a deterministic result provide a random number generator.

        Args:
            other (BSC): other input hypervector
            generator (``torch.Generator``, optional): a pseudorandom number generator for sampling.

        Shapes:
            - Self: :math:`(*)`
            - Other: :math:`(*)`
            - Output: :math:`(*)`

        Examples::

            >>> a, b = torchhd.BSC.random_hv(2, 10)
            >>> a
            tensor([ True, False,  True, False, False,  True,  True, False,  True, False])
            >>> b
            tensor([ True, False, False,  True,  True, False, False,  True, False,  True])
            >>> a.bundle(b)
            tensor([ True, False,  True,  True, False,  True,  True, False, False, False])

            >>> a, b = torchhd.BSC.random_hv(2, 10, dtype=torch.long)
            >>> a
            tensor([1, 0, 1, 1, 1, 0, 1, 1, 1, 1])
            >>> b
            tensor([1, 1, 1, 0, 0, 0, 1, 1, 0, 0])
            >>> a.bundle(b)
            tensor([1, 0, 1, 0, 0, 0, 1, 1, 0, 1])

        """

        tiebreaker = torch.empty_like(other)
        tiebreaker.bernoulli_(0.5, generator=generator)

        is_majority = self == other
        return self.where(is_majority, tiebreaker)

    def multibundle(self, *, generator: torch.Generator = None) -> "BSC":
        r"""Bundle multiple hypervectors."""
        if self.dim() < 2:
            class_name = self.__class__.__name__
            raise RuntimeError(
                f"{class_name} data needs to have at least two dimensions for multibind, got size: {tuple(self.shape)}"
            )

        n = self.size(-2)

        count = self.sum(dim=-2, dtype=torch.long)

        # add a tiebreaker when there are an even number of hvs
        if n % 2 == 0:
            tiebreaker = torch.empty_like(count)
            tiebreaker.bernoulli_(0.5, generator=generator)
            count += tiebreaker
            n += 1

        threshold = n // 2
        return torch.greater(count, threshold).to(self.dtype)

    def bind(self, other: "BSC") -> "BSC":
        r"""Bind the hypervector with other using XOR.

        This produces a hypervector dissimilar to both.

        Binding is used to associate information, for instance, to assign values to variables.

        Args:
            other (BSC): other input hypervector

        Shapes:
            - Self: :math:`(*)`
            - Other: :math:`(*)`
            - Output: :math:`(*)`

        Examples::

            >>> a, b = torchhd.BSC.random_hv(2, 10)
            >>> a
            tensor([ True, False,  True,  True, False,  True, False, False,  True, False])
            >>> b
            tensor([ True, False, False, False, False,  True, False, False, False, False])
            >>> a.bind(b)
            tensor([False, False,  True,  True, False, False, False, False,  True, False])

            >>> a, b = torchhd.BSC.random_hv(2, 10, dtype=torch.long)
            >>> a
            tensor([1, 0, 0, 1, 0, 1, 0, 0, 0, 0])
            >>> b
            tensor([0, 0, 0, 1, 0, 0, 1, 0, 1, 0])
            >>> a.bind(b)
            tensor([1, 0, 0, 0, 0, 1, 1, 0, 1, 0])

        """

        return self.logical_xor(other).to(other.dtype)

    def multibind(self) -> "BSC":
        """Bind multiple hypervectors."""
        if self.dim() < 2:
            class_name = self.__class__.__name__
            raise RuntimeError(
                f"{class_name} data needs to have at least two dimensions for multibind, got size: {tuple(self.shape)}"
            )

        n = self.size(-2)
        n_ = biggest_power_two(n)
        output = self[..., :n_, :]

        # parallelize many XORs in a hierarchical manner
        # for larger batches this is significantly faster
        while output.size(-2) > 1:
            output = torch.logical_xor(output[..., 0::2, :], output[..., 1::2, :])

        output = output.squeeze(-2)

        # TODO: as an optimization we could also perform the hierarchical XOR
        # on the leftovers in a recursive fashion
        leftovers = torch.unbind(self[..., n_:, :], -2)
        for i in range(n - n_):
            output = torch.logical_xor(output, leftovers[i])

        return output.to(self.dtype)

    def inverse(self) -> "BSC":
        r"""Invert the hypervector for binding.

        Each hypervector in BSC is its own inverse, so this returns a copy of self.

        Shapes:
            - Self: :math:`(*)`
            - Output: :math:`(*)`

        Examples::

            >>> a = torchhd.BSC.random_hv(1, 10)
            >>> a
            tensor([[False, False, False,  True,  True, False, False,  True,  True, False]])
            >>> a.inverse()
            tensor([[False, False, False,  True,  True, False, False,  True,  True, False]])

            >>> a = torchhd.BSC.random_hv(1, 10, dtype=torch.long)
            >>> a
            tensor([[0, 1, 0, 1, 1, 1, 1, 0, 1, 1]])
            >>> a.inverse()
            tensor([[0, 1, 0, 1, 1, 1, 1, 0, 1, 1]])

        """

        return self.clone()

    def negative(self) -> "BSC":
        r"""Negate the hypervector for the bundling inverse.

        Shapes:
            - Self: :math:`(*)`
            - Output: :math:`(*)`

        Examples::

            >>> a = torchhd.BSC.random_hv(1, 10)
            >>> a
            tensor([[ True,  True,  True,  True, False, False, False,  True,  True,  True]])
            >>> a.negative()
            tensor([[False, False, False, False,  True,  True,  True, False, False, False]])

            >>> a = torchhd.BSC.random_hv(1, 10, dtype=torch.long)
            >>> a
            tensor([[0, 1, 0, 1, 0, 0, 1, 1, 0, 1]])
            >>> a.negative()
            tensor([[1, 0, 1, 0, 1, 1, 0, 0, 1, 0]])

        """
        out = torch.empty_like(self).as_subclass(BSC)
        return torch.logical_not(self, out=out)

    def permute(self, shifts: int = 1) -> "BSC":
        r"""Permute the hypervector.

        The permutation operator is commonly used to assign an order to hypervectors.

        Args:
            shifts (int, optional): The number of places by which the elements of the tensor are shifted.

        Shapes:
            - Self: :math:`(*)`
            - Output: :math:`(*)`

        Examples::

            >>> a = torchhd.BSC.random_hv(1, 10)
            >>> a
            tensor([[ True, False, False, False, False, False, False, False, False, False]])
            >>> a.permute()
            tensor([[False,  True, False, False, False, False, False, False, False, False]])

            >>> a = torchhd.BSC.random_hv(1, 10, dtype=torch.long)
            >>> a
            tensor([[1, 1, 0, 0, 1, 1, 0, 0, 1, 1]])
            >>> a.permute()
            tensor([[1, 1, 1, 0, 0, 1, 1, 0, 0, 1]])

        """
        return self.roll(shifts=shifts, dims=-1)

    def dot_similarity(self, others: "BSC") -> Tensor:
        """Inner product with other hypervectors."""
        dtype = torch.get_default_dtype()

        min_one = torch.tensor(-1.0, dtype=dtype)
        plus_one = torch.tensor(1.0, dtype=dtype)

        self_as_bipolar = torch.where(self.bool(), min_one, plus_one)
        others_as_bipolar = torch.where(others.bool(), min_one, plus_one)

        return F.linear(self_as_bipolar, others_as_bipolar)

    def cos_similarity(self, others: "BSC") -> Tensor:
        """Cosine similarity with other hypervectors."""
        d = self.size(-1)
        return self.dot_similarity(others) / d
